/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2010, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: surface_mls.cpp 36648 2011-04-21 02:36:46Z michael.s.dixon $
 *
 */

/**
\author Radu Bogdan Rusu

@b surface_mls exemplifies how to use a MovingLeastSquares surface reconstruction algorithm.
**/
#include <boost/make_shared.hpp>
#include <ros/ros.h>
#include "pcl/io/io.h"
#include "pcl/point_types.h"
#include "pcl/surface/mls.h"
#include "pcl/kdtree/kdtree_flann.h"
#include "pcl_ros/publisher.h"

/* ---[ */
int
  main (int argc, char** argv)
{
  ros::init (argc, argv, "pcl_surface_mls_test");
  ros::NodeHandle nh;

  typedef pcl::PointXYZ Point;

  // ROS publisher
  pcl_ros::Publisher<Point> pub;
  pub.advertise (nh, "output", 1);

  typedef pcl::KdTree<Point>::Ptr KdTreePtr;
  KdTreePtr tree = boost::make_shared<pcl::KdTreeFLANN<Point> > ();
  
   // Datasets
  pcl::PointCloud<Point> cloud_raw;
  pcl::PointCloud<Point> cloud_mls;
  pcl::PointCloud<Point>::ConstPtr cloud_raw_ptr;
  // PCL Objects
  pcl::MovingLeastSquares<Point, pcl::PointNormal> mls;
  mls.setPolynomialFit (true);
  mls.setSearchMethod (tree);
  mls.setSearchRadius (0.02);

  // Spin
  while (nh.ok ())
  {
    // Wait for message
    sensor_msgs::PointCloud2::ConstPtr cloud = ros::topic::waitForMessage<sensor_msgs::PointCloud2>("input");
    ROS_INFO ("PointCloud with %d data points (%s), stamp %f, and frame %s on topic %s received.",
              cloud->width * cloud->height, pcl::getFieldsList (*cloud).c_str (), cloud->header.stamp.toSec (), cloud->header.frame_id.c_str (), nh.resolveName ("input").c_str ());
  
    // Convert to PointCloud<T>
    pcl::fromROSMsg (*cloud, cloud_raw);
    cloud_raw_ptr.reset (new pcl::PointCloud<Point> (cloud_raw));

    // Create a MLS representation
    mls.setInputCloud (cloud_raw_ptr);
    mls.reconstruct (cloud_mls);

    // Publish
    pub.publish (cloud_mls);
  }

  return (0);
}
/* ]--- */
